/*

    ****************************************************
    *   Computation of a fractionnal Brownian motion   *
    *   using the circulant matrix embedding method    *
    ****************************************************

    Copyright (C) 2003 Samuel Nicolay.
    email: <S.Nicolay@ulg.ac.be>


    This program is a free software, you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program (in a file named COPYRIGHT);
    if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
To compile, type something like 
gcc circul.c random.c utils.c -lfftw -lm -o fbm_circ
*/

#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include<fftw.h>
#include<time.h>
#include"random.h" /*This generates a white noise*/
#include"utils.h"  /*Some useful easy functions*/

void help(char *file_name);
void calc_fbm(fftw_complex *x, double *z, double size, double h);

int main(int arg_c,char **arg_v){
  int j,no_name=0;
  long int size,coeff;
  double *tab;
  float sum,h;
  char *name;
  fftw_complex *output;
  FILE *output_file;

/*
************************************
*   First, we get the parameters   *
************************************
*/

  for (j=1;j<arg_c;j++) 
    if (arg_v[j][0]=='-')
      if (w1_is_w2("help",arg_v[j]+1,'-')){help(arg_v[0]);exit(0);}

  if (arg_c<2) error("Hurst exponent not specified.\nTry the option '--help'.");
  else {
    if (string_is_float(arg_v[1])) h=(float) string2float(arg_v[1]);
    else error("The first argument must be the Hurst exponent.");
    if(h>=1 || h<0) error("The Hurst exponent H must be such that 0<H<1.");
  }

  if (arg_c<3) error("Size not specified.");
  else{
    if (string_is_int(arg_v[2])) size=(long int) string2float(arg_v[2]);
    else error("The second argument must be the size.");
  }
  if (arg_c<4) {
    name = (char*) malloc(8*sizeof(char));
    name[0]='f';name[1]='b';name[2]='m';name[3]='.';name[4]='d';name[5]='a';name[6]='t';name[7]='\0';
    no_name=1;
    fprintf(stderr,"Output filename not specified.\nThe default name is \"%s\".\n",name);
  }
  else name=arg_v[3];

  if (arg_c<5) coeff=2;
  else {
    if(string_is_int(arg_v[4])) coeff=(long int) string2float(arg_v[4]);
    else coeff=2;
  }



/*
********************************
*   Now, we generate the fBm   *
********************************
*/

  tab = (double*) malloc(coeff*(size-1)*sizeof(double));

  output=(fftw_complex*) malloc(coeff*(size-1)*sizeof(fftw_complex));

  for(j=0;j<coeff*(size-1);j++)tab[j]=Grand(1.0);

  calc_fbm(output,tab,coeff*(size-1),h);

/*
**********************
*   And we save it   *
**********************
*/

  if ((output_file=fopen(name,"w"))==NULL) fprintf(stderr,"Could not open \"%s\" for writing.\n",name);
  sum=0;
  for(j=0;j<size;j++){sum+=output[j].re;fprintf(output_file,"%f\n",sum);}
  fclose(output_file);

/*
*******************************************
*   Time to end, so we free the vectors   *
*******************************************
*/

  free(tab);
  free(output);
  if (no_name) free(name);

  return 0;
}



void help(char *file_name){
/*
*********************
*   Give the help   *
*********************
*/
  printf("%s H size [output_filename] [size factor]\n",file_name);
  printf("Generate exact fractional Brownian motion of size \"size\" with Hurst exponent \"H\" using the circulant matrix embedding method.\n");
  printf("The output will be saved in the file \"output_filename\".\n");
  printf("To increase the working size, one can enter a \"size factor\" greater than 2; anyway this argument is unuseful here.\n");
}



void calc_fbm(fftw_complex *x, double *z, double size, double h){
/*
********************************************
*   Compute a fractional Brownian motion   *
********************************************
*/
  void calc_w(fftw_complex *w,double *z,double size,double h);
  fftw_complex *w;
  fftw_plan p;

  w=(fftw_complex*) malloc(size*sizeof(fftw_complex));
  calc_w(w,z,size,h);

  p=fftw_create_plan(size,FFTW_BACKWARD,FFTW_ESTIMATE);
  fftw_one(p,w,x);

  fftw_destroy_plan(p);

  free(w);

}



void calc_w(fftw_complex *w,double *z,double size,double h){
/*
*********************************
*   Computation of the Matrix   *
*********************************
*/
  void calc_ev(fftw_complex *lamb,int size,double h);
  int j;
  fftw_complex *cz,*ev;
  fftw_plan p;

  ev=(fftw_complex*) malloc(size*sizeof(fftw_complex));

  calc_ev(ev,size,h);

  cz=(fftw_complex*) malloc(size*sizeof(fftw_complex));
  for(j=0;j<size;j++){cz[j].re=z[j];cz[j].im=0;}

  p=fftw_create_plan(size,FFTW_FORWARD,FFTW_ESTIMATE);
  fftw_one(p,cz,w);
  for(j=0;j<size;j++){w[j].re /= size; w[j].im /= size;} /*should be sqrt(size)*/
  for(j=0;j<size;j++){w[j].re *= sqrt(ev[j].re); w[j].im *= sqrt(ev[j].re);}

  fftw_destroy_plan(p);
  free(ev);
  free(cz);
}



void calc_ev(fftw_complex *lamb,int size,double h){
/*
***************************************
*   Computation of the Eigen Values   *
***************************************
*/
  double r_j(double j,double h);
  int j;
  fftw_complex *mat;
  fftw_plan p;

  mat=(fftw_complex*) malloc(size*sizeof(fftw_complex));
  for (j=0;j<size;j++){
    //    if(j<=size/2)mat[j].re=r_j(j,h);
    if(j<=(size>>1))mat[j].re=r_j(j,h);
    else mat[j].re=r_j(size-j,h);
    mat[j].im=0;
  }

  p=fftw_create_plan(size,FFTW_BACKWARD,FFTW_ESTIMATE);
  fftw_one(p,mat,lamb);

  for (j=0;j<size;j++) if (lamb[j].re<0||fabs(lamb[j].im)>10e-3) fprintf(stderr,"The %d-th eigen value (%f + i%f) is not positive.\n",j,lamb[j].re,lamb[j].im);

  fftw_destroy_plan(p);
  free(mat);
}



double r_j(double j,double h){
/*
************************************
*   The autocorrelation function   *
************************************
*/
  double retour;

  retour=pow(fabs(j+1),2*h)+pow(fabs(j-1),2*h)-2*pow(fabs(j),2*h);
  retour *= .5; /*Here, the variance is equal to 1*/

  return retour;
}
