/*

    ****************************************************
    *   Computation of a fractionnal Brownian motion   *
    *              using the farima method             *
    ****************************************************

    Copyright (C) 2003 Samuel Nicolay.
    email: <S.Nicolay@ulg.ac.be>


    This program is a free software, you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program (in a file named COPYRIGHT);
    if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
To compile, type something like 
gcc farima.c random.c utils.c -lfftw -lm -o fbm_farima
*/
#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include<fftw.h>
#include<time.h>
#include"random.h" /*This generates a white noise*/
#include"utils.h"  /*Some useful easy functions*/

void help(char *file_name);
static double Ahi(double n,int k);
void convol(float *coef,float *noise,float *output,long int size);


int main(int arg_c,char **arg_v){
  int no_name;
  long int j,size,border;
  float h,*coef,*noise,*output;
  char *name;
  FILE *output_file;

/*
************************************
*   First, we get the parameters   *
************************************
*/

  for (j=1;j<arg_c;j++) 
    if (arg_v[j][0]=='-')
      if (w1_is_w2("help",arg_v[j]+1,'-')){help(arg_v[0]);exit(0);}

  if (arg_c<2) error("Hurst exponent not specified.\nTry the option '--help'.");
  else {
    if (string_is_float(arg_v[1])) h=(float) string2float(arg_v[1]);
    else error("The first argument must be the Hurst exponent.");
    if(h>=1 || h<0) error("The Hurst exponent H must be such that 0<H<1.");
  }

  if (arg_c<3) error("Size not specified.");
  else{
    if (string_is_int(arg_v[2])) size=(long int) string2float(arg_v[2]);
    else error("The second argument must be the size.");
  }
  if (arg_c<4) {
    name = (char*) malloc(8*sizeof(char));
    name[0]='f';name[1]='b';name[2]='m';name[3]='.';name[4]='d';name[5]='a';name[6]='t';name[7]='\0';
    no_name=1;
    fprintf(stderr,"Output filename not specified.\nThe default name is \"%s\"\n",name);
  }
  else name=arg_v[3];

  if (arg_c<5) border=10;
  else {
    if(string_is_int(arg_v[4])) border=(long int) string2float(arg_v[4]);
    else border=10;
  }

  //printf("H=%f Taille: %ld ->\"%s\" (%d).\n",h,size,name,border);


/*
********************************
*   Now, we generate the fBm   *
********************************
*/

  size += border;
  coef = (float*) malloc((size<<1)*sizeof(float));
  for(j=0;j<(size<<1);j++)coef[j]=0;  
  noise = (float*) malloc((size<<1)*sizeof(float));
  for(j=0;j<(size<<1);j++)noise[j]=0;  
  output = (float*) malloc(size*sizeof(float));
  for(j=0;j<size;j++)output[j]=0;  


  /*for(j=0;j<size;j++)coef[j]=j;
  for(j=0;j<size;j++)noise[j]=1;*/
  for(j=0;j<size;j++)coef[j]=Ahi(h,j);
  for(j=0;j<size;j++)noise[j]=Grand(1.0);
  convol(coef,noise,output,size);
//  for(j=0;j<size;j++)pwet[j]=coef[j]*noise[size-j-1];

/*
**********************
*   And we save it   *
**********************
*/

  size -= border;
  if ((output_file=fopen(name,"w"))==NULL) fprintf(stderr,"Could not open \"%s\" for writing.\n",name);
  for(j=0;j<size;j++){
    fprintf(output_file,"%f\n",output[j]);
  }
  fclose(output_file);

/*
*******************************************
*   Time to end, so we free the vectors   *
*******************************************
*/

  if (no_name) free(name);
  free(coef);
  free(noise);
  free(output);

  return 0;
}



void help(char *file_name){
/*
*********************
*   Give the help   *
*********************
*/
  printf("%s H size [output_filename] [border size]\n",file_name);
  printf("Generate fractional Brownian motion of size \"size\" with Hurst exponent \"H\" using the farima method.\n");
  printf("The output will be saved in the file \"output_filename\".\n");
  printf("One may change the \"border size\".\nBy default this size is 10.\n");
}



static double LnGamma(double xx,int *psign){
/*
******************************************************
*   Useful for the function Ahi (give Ln of Gamma)   *
******************************************************
*/
  double x, tmp, ser;
  static double 
    cof[6]={76.18009173,-86.50532033,24.01409822,-1.231739516,0.120858003e-2,-0.536382e-5}; 
  
  int j;
  *psign=1;
  if (xx<=0){
    if (sin(pi*xx)<0) *psign=-1;
  return(log(pi)-LnGamma(1-xx,&j)-log(fabs(sin(pi*xx))));
  }
  
  x=xx-1.0;
  tmp=x+5.5;
  tmp-=(x+0.5)*log(tmp);
  ser=1.0;
  for(j=0;j<=5;j++){
    x+=1.0;
    ser+=cof[j]/x;
  }
  return -tmp+log(2.50662827465*ser);
}



static double Ahi(double n,int k){
/*
***************************************************
*   Give the numerical value of the coefficient   *
***************************************************
*/
  static double LnGamma(double xx,int *psign);
  int sign1,sign2,sign3;
  double res;
  
  res=exp(LnGamma(n+k+0.5,&sign1)-LnGamma(n+0.5,&sign2)-LnGamma(k+1,&sign3));
  res *= sign1*sign2*sign3;
  return(res);
}



void convol(float *coef,float *noise,float *output,long int size){
/*
*************************************
*   Make the discrete convolution   *
*************************************
*/
  void calc_prod(fftw_complex *prod1,fftw_complex *prod2,long int size);
  void fft_conv(float *tableau,fftw_complex*,long int size);
  void fftI_conv(fftw_complex*,fftw_complex*,long int size);
  
  long int j;
  fftw_complex *output_fft1,*output_fft2;

  output_fft1=(fftw_complex*) malloc((size<<1)*sizeof(fftw_complex));
  output_fft2=(fftw_complex*) malloc((size<<1)*sizeof(fftw_complex));

  fft_conv(coef,output_fft1,size);
  fft_conv(noise,output_fft2,size);
  calc_prod(output_fft1,output_fft2,size);
  fftI_conv(output_fft1,output_fft2,size);
  for(j=0;j<size;j++)output[j]=output_fft2[j].re/(size<<1);

  free(output_fft1);
  free(output_fft2);
}



void calc_prod(fftw_complex *prod1,fftw_complex *prod2,long int size){
/*
*****************************************************
*   Compute the product between two complex numbers   *
*****************************************************
*/
  long int j;
  float re,im;

  for(j=0;j<size<<1;j++){
    re=prod1[j].re*prod2[j].re-prod1[j].im*prod2[j].im;
    im=prod1[j].re*prod2[j].im+prod1[j].im*prod2[j].re;
    prod1[j].re=re;
    prod1[j].im=im;
  }
}



void fft_conv(float *tableau,fftw_complex *out,long int size){
/*
*********************************
*   Make the fft with padding   *
*********************************
*/
  int j;

  fftw_complex *in;
  fftw_plan p;

  in=(fftw_complex*) malloc((size<<1)*sizeof(fftw_complex));
  for(j=0;j<(size<<1);j++){in[j].re=tableau[j];in[j].im=0;}

  p=fftw_create_plan(size<<1,FFTW_FORWARD,FFTW_ESTIMATE);
  fftw_one(p,in,out);

  fftw_destroy_plan(p);
  free(in);
}



void fftI_conv(fftw_complex *in,fftw_complex *out,long int size){
/*
*****************************************
*   Make the inverse fft with padding   *
*****************************************
*/
  fftw_plan p;

  p=fftw_create_plan(size<<1,FFTW_BACKWARD,FFTW_ESTIMATE);
  fftw_one(p,in,out);

  fftw_destroy_plan(p);
}
