/*

    ****************************************************
    *   Computation of a fractionnal Brownian motion   *
    *             using a spectral method              *
    ****************************************************

    Copyright (C) 2003 Samuel Nicolay.
    email: <S.Nicolay@ulg.ac.be>


    This program is a free software, you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program (in a file named COPYRIGHT);
    if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
To compile, type something like 
gcc spectral.c random.c utils.c -lfftw -lm -o fbm_spec
*/

#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include<fftw.h>
#include"utils.h"
#include"random.h"

void help(char *file_name);
void compute_bk(float *gauss1,float *gauss2, float *output, long int size,double h,double prec);
int main(int arg_c, char **arg_v){
  int no_name=0;
  float *gauss1,*gauss2,*output;
  long int j,size,real_size;
  float sum,h;
  double prec;
  char *name;
  FILE *output_file;

/*
************************************
*   First, we get the parameters   *
************************************
*/

  for (j=1;j<arg_c;j++) 
    if (arg_v[j][0]=='-')
      if (w1_is_w2("help",arg_v[j]+1,'-')){help(arg_v[0]);exit(0);}

  if (arg_c<2) error("Hurst exponent not specified.\nTry the option '--help'.");
  else {
    if (string_is_float(arg_v[1])) h=(float) string2float(arg_v[1]);
    else error("The first argument must be the Hurst exponent.");
    if(h>=1 || h<0) error("The Hurst exponent H must be such that 0<H<1.");
  }

  if (arg_c<3) error("Size not specified.");
  else{
    if (string_is_int(arg_v[2])) size=(long int) string2float(arg_v[2]);
    else error("The second argument must be the size.");
  }
  if (arg_c<4) {
    name = (char*) malloc(8*sizeof(char));
    name[0]='f';name[1]='b';name[2]='m';name[3]='.';name[4]='d';name[5]='a';name[6]='t';name[7]='\0';
    no_name=1;
    fprintf(stderr,"Output filename not specified.\nThe default name is \"%s\"\n",name);
  }
  else name=arg_v[3];

  if (arg_c<5) prec=0.01;
  else {
    if(string_is_float(arg_v[4])) prec=string2float(arg_v[4]);
    else prec=0.01;
  }


/*
********************************
*   Now, we generate the fBm   *
********************************
*/


  real_size=size;
  if ((size&1)==1)real_size++;

  gauss1=(float*) malloc(real_size*sizeof(float));
  gauss2=(float*) malloc(real_size*sizeof(float));
  output=(float*) malloc(real_size*sizeof(float));
  for(j=0;j<real_size;j++){gauss1[j]=Grand(1.0);gauss2[j]=Grand(1.0);}
  compute_bk(gauss1,gauss2,output,real_size,h,prec);

/*
**********************
*   And we save it   *
**********************
*/

  if ((output_file=fopen(name,"w"))==NULL) fprintf(stderr,"Could not open \"%s\" for writing.\n",name);
  sum=0;
  for(j=0;j<size;j++){sum+=output[j];fprintf(output_file,"%f\n",sum);}
//  for(j=0;j<size;j++){fprintf(output_file,"%f\n",output[j].re);}
  fclose(output_file);

/*
*******************************************
*   Time to end, so we free the vectors   *
*******************************************
*/

  if(no_name) free(name);
  free(gauss1);
  free(gauss2);
  free(output);

  return 0;
}

void compute_bk(float *gauss1,float *gauss2, float *output, long int size,double h,double prec){
/*
***********************
*   Compute the fBm   *
***********************
*/
  double fctF (double lam,double size,double h, double prec);

  fftw_complex *bk,*resul;
  float *lam;
  fftw_plan p;
  long int j;

  bk=(fftw_complex*) malloc(size*sizeof(fftw_complex));
  resul=(fftw_complex*) malloc(size*sizeof(fftw_complex));

  lam=(float*) malloc(size*sizeof(float));
  for(j=0;j<size;j++)lam[j]=2*pi*j/size;

  bk[0].re=0;bk[0].im=0;
  for(j=1;j<=(size>>1);j++){
    bk[j].re=gauss1[j]*sqrt(fctF(lam[j],size,h,prec)/(size<<1));
    bk[j].im=gauss2[j]*sqrt(fctF(lam[j],size,h,prec)/(size<<1));
  }

  for(j=(size>>1)+1;j<size;j++){
    bk[j].re=gauss1[size-j]*sqrt(fctF(lam[j],size,h,prec)/(size<<1));
    bk[j].im=-gauss2[size-j]*sqrt(fctF(lam[j],size,h,prec)/(size<<1));
  }


  p=fftw_create_plan(size,FFTW_FORWARD,FFTW_ESTIMATE);
  fftw_one(p,bk,resul);

  //for(j=0;j<size>>1;j++)printf("%f + %fi\n",resul[j].re,resul[j].im);

  for(j=0;j<size;j++)output[j]=resul[j].re;

  fftw_destroy_plan(p);
  free(bk);
  free(resul);
  free(lam);
}


void help(char *file_name){
/*
*********************
*   Give the help   *
*********************
*/
  printf("%s H size [output_filename] [precision]\n",file_name);
  printf("Generate fractional Brownian motion of size \"size\" with Hurst exponent \"H\" using a spectral method.\n");
  printf("The output will be saved in the file \"output_filename\".\n");
  printf("One can increase the precision of the approximation, increasing \"precision\".\nBy default it is set to 0.001.\n");
}



double fctF (double lam,double size,double h,double prec){
/*
**********************
*   Give f(lambda)   *
**********************
*/
  static double LnGamma(double xx,int *psign);
  double fctB(double lam,double h,double prec);

  int sign;
  double result;

  result= 2*sin(pi*h)*exp(LnGamma(2*h+1,&sign))*(1-cos(lam))*(pow(fabs(lam),-2*h-1)+fctB(lam,h,prec));
  
  return ((double) result*sign);
}

double fctB(double lam,double h,double prec){
/*
**********************
*   Give B(lambda)   *
**********************
*/
  int j;
  double sum, prev_sum;

  sum=0;j=1;
  do{
    prev_sum=sum;
    sum += pow(2*pi*j+lam,-2*h-1)+ pow(2*pi*j-lam,-2*h-1);
    j++;
  }while (fabs(sum-prev_sum)>prec);
  return sum;
}



static double LnGamma(double xx,int *psign){
/*
*************************
*   Give Ln of Gamma   *
*************************
*/
  double x, tmp, ser;
  static double 
    cof[6]={76.18009173,-86.50532033,24.01409822,-1.231739516,0.120858003e-2,-0.536382e-5}; 
  
  int j;
  *psign=1;
  if (xx<=0){
    if (sin(pi*xx)<0) *psign=-1;
  return(log(pi)-LnGamma(1-xx,&j)-log(fabs(sin(pi*xx))));
  }
  
  x=xx-1.0;
  tmp=x+5.5;
  tmp-=(x+0.5)*log(tmp);
  ser=1.0;
  for(j=0;j<=5;j++){
    x+=1.0;
    ser+=cof[j]/x;
  }
  return -tmp+log(2.50662827465*ser);
}
